# -*- coding: utf-8 -*-
# ###
# Copyright (c) 2015, Rice University
# This software is subject to the provisions of the GNU Affero General
# Public License version 3 (AGPLv3).
# See LICENCE.txt for details.
# ###
"""Generate a migration script in the migrations directory."""

import os

from ..utils import timestamp
from .. import logger


__all__ = ('cli_loader',)


def cli_command(migration_name='', **kwargs):
    filename = '{}_{}.py'.format(timestamp(), migration_name)
    directory = kwargs['migrations_directory']
    if not directory:
        raise Exception('migrations directory undefined')
    if len(directory) > 1:
        raise Exception('more than one migrations directory specified')
    directory = directory[0]
    path = os.path.join(directory, filename)
    if not os.path.isdir(directory):
        os.makedirs(directory)
    with open(path, 'w') as f:
        f.write("""\
# -*- coding: utf-8 -*-


# Uncomment should_run if this is a repeat migration
# def should_run(cursor):
#     # TODO return True if migration should run


def up(cursor):
    # TODO migration code
    pass

    # if a super user database connection is needed
    # from dbmigrator import super_user
    # with super_user() as super_cursor:
    #     pass


def down(cursor):
    # TODO rollback code
    pass
""")
    logger.info('Generated migration script "{}"'.format(path))


def cli_loader(parser):
    parser.add_argument('migration_name')
    return cli_command



# -*- coding: utf-8 -*-
# ###
# Copyright (c) 2016, Rice University
# This software is subject to the provisions of the GNU Affero General
# Public License version 3 (AGPLv3).
# See LICENCE.txt for details.
# ###
"""Mark a migration as completed or not completed."""

from .. import logger, utils


__all__ = ('cli_loader',)


@utils.with_cursor
def cli_command(cursor, migrations_directory='', migration_timestamps=None,
                completed=None, **kwargs):
    if completed is None:
        raise Exception('-t, -f or -d must be supplied.')

    migrations = {version: migration
                  for version, _, migration in utils.get_migrations(
                      migrations_directory, import_modules=True)}

    for migration_timestamp in migration_timestamps:
        migration = migrations.get(migration_timestamp)
        if migration is None:
            raise SystemExit(
                'Migration {} not found'.format(migration_timestamp))

        utils.mark_migration(cursor, migration_timestamp, completed)
        if not completed:
            message = 'not been run and not deferred'
        elif completed == 'deferred':
            message = 'deferred'
        else:
            message = 'completed'
        logger.info('Migration {} marked as {}'.format(migration_timestamp,
                                                       message))


def cli_loader(parser):
    parser.add_argument('migration_timestamps', nargs='+')
    parser.add_argument('-t', action='store_const',
                        const=True,
                        dest='completed',
                        default=None,
                        help='Mark the migration as completed')
    parser.add_argument('-f', action='store_const',
                        const=False,
                        dest='completed',
                        default=None,
                        help='Mark the migration as not been run')
    parser.add_argument('-d', action='store_const',
                        const='deferred',
                        dest='completed',
                        default=None,
                        help='Mark the migration as deferred')
    return cli_command



# -*- coding: utf-8 -*-
# ###
# Copyright (c) 2015, Rice University
# This software is subject to the provisions of the GNU Affero General
# Public License version 3 (AGPLv3).
# See LICENCE.txt for details.
# ###
"""Initialize schema migrations table."""

from .. import logger, utils


__all__ = ('cli_loader',)


@utils.with_cursor
def cli_command(cursor, migrations_directory='', version=None, **kwargs):
    cursor.execute("""\
        SELECT 1 FROM information_schema.tables
        WHERE table_name = 'schema_migrations'""")
    table_exists = cursor.fetchone()
    if table_exists:
        logger.info('Schema migrations already initialized.')
        return

    cursor.execute("""\
        CREATE TABLE schema_migrations (
            version TEXT NOT NULL,
            applied TIMESTAMP WITH TIME ZONE DEFAULT CURRENT_TIMESTAMP
        )""")
    versions = []
    if version is None:
        timestamp = utils.timestamp()
    else:
        timestamp = str(version)
    for version, name in utils.get_migrations(migrations_directory):
        if version <= timestamp:
            versions.append((version,))
    cursor.executemany("""\
        INSERT INTO schema_migrations VALUES (%s)
        """, versions)
    logger.info('Schema migrations initialized.')


def cli_loader(parser):
    parser.add_argument('--version', type=int,
                        help='Set the schema version to VERSION, '
                             'default current timestamp')
    return cli_command



# -*- coding: utf-8 -*-
# ###
# Copyright (c) 2015, Rice University
# This software is subject to the provisions of the GNU Affero General
# Public License version 3 (AGPLv3).
# See LICENCE.txt for details.
# ###
"""Rollback a migration."""

from .. import logger, utils


__all__ = ('cli_loader',)


@utils.with_cursor
def cli_command(cursor, migrations_directory='', steps=1,
                db_connection_string='', **kwargs):
    migrated_versions = list(utils.get_schema_versions(
        cursor, include_deferred=False, order_by='applied'))
    logger.debug('migrated_versions: {}'.format(migrated_versions))
    if not migrated_versions:
        logger.info('No migrations to roll back.')
        return
    migrations = {version: (name, migration)
                  for version, name, migration in utils.get_migrations(
                      migrations_directory, import_modules=True,
                      reverse=True)}

    rolled_back = 0
    for version in reversed(migrated_versions):
        if version not in migrations:
            logger.info('Migration {} not found.'.format(version))
            break
        migration_name, migration = migrations[version]
        utils.compare_schema(db_connection_string,
                             utils.rollback_migration,
                             cursor,
                             version,
                             migration_name,
                             migration)
        rolled_back += 1
        if rolled_back >= steps:
            break

    if not rolled_back:
        logger.info('No migrations to roll back.')


def cli_loader(parser):
    parser.add_argument('--steps', metavar='N', default=1, type=int,
                        help='Roll back the last N migrations, default 1')
    return cli_command



# -*- coding: utf-8 -*-
# ###
# Copyright (c) 2015, Rice University
# This software is subject to the provisions of the GNU Affero General
# Public License version 3 (AGPLv3).
# See LICENCE.txt for details.
# ###

import os
import re
import sys
from setuptools import setup, find_packages

install_requires = (
    'psycopg2>=2.7',
    )

tests_require = [
    ]

if sys.version_info.major < 3:
    tests_require.append('mock')

here = os.path.dirname(__file__)


def read(path):
    with open(os.path.join(here, path)) as f:
        return f.read()


def find_version(path):
    m = re.search(r'^__version__\s*=\s*[\'"]([^\'"]*)[\'"]',
                  read(path), re.MULTILINE)
    if m:
        return m.group(1)
    raise RuntimeError('Unable to find version string')

LONG_DESC = '\n\n~~~~\n\n'.join([read('README.rst'),
                                 read('CHANGELOG.rst')])

setup(
    name='db-migrator',
    version=find_version('dbmigrator/__init__.py'),
    author='Connexions',
    author_email='info@cnx.org',
    url='https://github.com/karenc/db-migrator',
    license='AGPL, see also LICENSE.txt',
    description='Python package to migrate postgresql database',
    long_description=LONG_DESC,
    packages=find_packages(),
    install_requires=install_requires,
    tests_require=tests_require,
    test_suite='dbmigrator.tests',
    include_package_data=True,
    entry_points={
        'console_scripts': [
            'dbmigrator = dbmigrator.cli:main',
            ],
        },
    )



# -*- coding: utf-8 -*-
# ###
# Copyright (c) 2016, Rice University
# This software is subject to the provisions of the GNU Affero General
# Public License version 3 (AGPLv3).
# See LICENCE.txt for details.
# ###

import datetime
import os.path
import tempfile
import unittest
try:
    from unittest import mock
except ImportError:
    import mock

import psycopg2

from . import testing
from ..utils import db_connect


logger = mock.Mock()


class UtilsTestCase(unittest.TestCase):
    def setUp(self):
        import dbmigrator
        _logger = dbmigrator.utils.logger
        self.addCleanup(setattr, dbmigrator.utils, 'logger', _logger)
        dbmigrator.utils.logger = logger
        with db_connect(testing.db_connection_string) as db_conn:
            with db_conn.cursor() as cursor:
                cursor.execute("""
CREATE TABLE schema_migrations (
    version TEXT NOT NULL,
    applied TIMESTAMP WITH TIME ZONE DEFAULT CURRENT_TIMESTAMP
)""")
        dbmigrator.utils.set_settings({
            'db_connection_string': testing.db_connection_string,
        })

    def tearDown(self):
        with db_connect(testing.db_connection_string) as db_conn:
            with db_conn.cursor() as cursor:
                cursor.execute('DROP TABLE IF EXISTS a_table')
                cursor.execute('DROP TABLE IF EXISTS schema_migrations')

    def test_super_user(self):
        from ..utils import super_user, with_cursor

        def cleanup():
            with db_connect(testing.db_connection_string,
                            user='postgres') as db_conn:
                with db_conn.cursor() as cursor:
                    cursor.execute('DROP TABLE IF EXISTS super_table')

        self.addCleanup(cleanup)

        with db_connect(testing.db_connection_string) as db_conn:
            with db_conn.cursor() as cursor:
                cursor.execute('SELECT current_user')
                self.assertEqual('travis', cursor.fetchone()[0])

                with super_user() as super_cursor:
                    super_cursor.execute('SELECT current_user')
                    self.assertEqual('postgres', super_cursor.fetchone()[0])

                    super_cursor.execute('CREATE TABLE super_table (id TEXT)')
                    super_cursor.execute(
                        'GRANT ALL PRIVILEGES ON TABLE super_table TO travis')
                    super_cursor.execute(
                        "INSERT INTO super_table VALUES ('62')")

                cursor.execute('SELECT current_user')
                self.assertEqual('travis', cursor.fetchone()[0])

                cursor.execute('SELECT * FROM super_table')
                self.assertEqual('62', cursor.fetchone()[0])

    def test_get_settings_from_entry_points(self):
        from ..utils import get_settings_from_entry_points

        testing.install_test_packages()

        contexts = testing.test_packages
        settings = {
            'db_connection_string': testing.db_connection_string,
            }

        get_settings_from_entry_points(settings, contexts)

        self.assertEqual(
            settings,
            {'migrations_directory': testing.test_migrations_directories,
             'db_connection_string': testing.db_connection_string,
             })

    def test_get_settings_from_config(self):
        from ..utils import get_settings_from_config

        tmp_dir = tempfile.gettempdir()

        settings = {
            'migrations_directory': tmp_dir,
            }

        get_settings_from_config(
            testing.test_config_path,
            ['db-connection-string', 'migrations-directory', 'does-not-exist'],
            settings)

        self.assertEqual(
            settings,
            {'db_connection_string': testing.db_connection_string,
             'migrations_directory': tmp_dir})

    def test_with_cursor(self):
        from ..utils import with_cursor

        self.called = False

        @with_cursor
        def func(cursor, arg_1, kwarg_1='kwarg_1', kwarg_2='kwarg_2',
                 db_connection_string=None):
            self.assertTrue(isinstance(cursor, psycopg2.extensions.cursor))
            self.assertEqual(arg_1, 'arg_1')
            self.assertEqual(kwarg_1, 'called')
            self.assertEqual(kwarg_2, 'kwarg_2')
            self.assertEqual(
                db_connection_string, testing.db_connection_string)
            self.called = True

        func('arg_1', kwarg_1='called',
             db_connection_string=testing.db_connection_string)
        self.assertTrue(self.called)

        with self.assertRaises(Exception) as cm:
            func('')

        self.assertEqual(str(cm.exception), 'db-connection-string missing')

    def test_import_migration(self):
        from ..utils import import_migration

        migration_path = os.path.join(
            testing.test_data_path, 'package-a', 'package_a', 'migrations',
            '20160228202637_add_table.py')

        migration = import_migration(migration_path)
        self.assertIn('up', dir(migration))
        self.assertIn('down', dir(migration))

        with db_connect(testing.db_connection_string) as db_conn:
            with db_conn.cursor() as cursor:
                try:
                    cursor.execute('SELECT * FROM a_table')
                    self.fail('a_table should not be in the database before '
                              'the migration is run')
                except psycopg2.ProgrammingError as e:
                    self.assertIn('relation "a_table" does not exist', str(e))

        with db_connect(testing.db_connection_string) as db_conn:
            with db_conn.cursor() as cursor:
                # Run the migration.
                migration.up(cursor)

                # Try accessing a_table created by the migration.
                cursor.execute('SELECT * FROM a_table')

                # Roll back the migration.
                migration.down(cursor)

        with db_connect(testing.db_connection_string) as db_conn:
            with db_conn.cursor() as cursor:
                try:
                    cursor.execute('SELECT * FROM a_table')
                    self.fail('a_table should not be in the database after '
                              'the migration is rolled back')
                except psycopg2.ProgrammingError as e:
                    self.assertIn('relation "a_table" does not exist', str(e))

    def test_get_migrations(self):
        from ..utils import get_migrations

        migrations = get_migrations(testing.test_migrations_directories)

        self.assertEqual(
            list(migrations),
            [('20160228202637', 'add_table'),
             ('20160228210326', 'initial_data'),
             ('20160228212456', 'cool_stuff')])

    def test_get_migrations_no_migrations_directories(self):
        from ..utils import get_migrations

        migrations = get_migrations([])

        self.assertEqual(list(migrations), [])

    def test_get_pending_migrations(self):
        from ..utils import get_pending_migrations

        with db_connect(testing.db_connection_string) as db_conn:
            with db_conn.cursor() as cursor:
                migrations = list(get_pending_migrations(
                    testing.test_migrations_directories, cursor))

        self.assertEqual(migrations, [
            ('20160228202637', 'add_table'),
            ('20160228210326', 'initial_data'),
            ('20160228212456', 'cool_stuff')])

    def test_run_migration(self):
        from ..utils import run_migration, get_pending_migrations

        with db_connect(testing.db_connection_string) as db_conn:
            with db_conn.cursor() as cursor:
                migrations = list(get_pending_migrations(
                    testing.test_migrations_directories, cursor,
                    import_modules=True))
                version, name, migration = list(migrations)[0]

                run_migration(cursor, version, name, migration)

                after_migrations = list(get_pending_migrations(
                    testing.test_migrations_directories, cursor))

        logger.info.assert_called_with(
            'Running migration {} {}'.format(version, name))

        self.assertNotIn((version, name), after_migrations)

    def test_rollback_migration(self):
        from ..utils import (
            rollback_migration, run_migration, get_pending_migrations)

        with db_connect(testing.db_connection_string) as db_conn:
            with db_conn.cursor() as cursor:
                migrations = list(get_pending_migrations(
                    testing.test_migrations_directories, cursor,
                    import_modules=True))
                version, name, migration = list(migrations)[0]

                run_migration(cursor, version, name, migration)

                rollback_migration(cursor, version, name, migration)

                after_migrations = list(get_pending_migrations(
                    testing.test_migrations_directories, cursor))

        logger.info.assert_called_with(
            'Rolling back migration {} {}'.format(version, name))

        self.assertIn((version, name), after_migrations)

    def test_timestamp(self):
        from ..utils import timestamp

        now = datetime.datetime(2016, 2, 28, 22, 51, 56)

        with mock.patch('datetime.datetime') as mock_datetime:
            mock_datetime.utcnow.return_value = now
            self.assertEqual(timestamp(), '20160228225156')



# -*- coding: utf-8 -*-
# ###
# Copyright (c) 2015, Rice University
# This software is subject to the provisions of the GNU Affero General
# Public License version 3 (AGPLv3).
# See LICENCE.txt for details.
# ###

try:
    import configparser
except ImportError:
    # python 2
    import ConfigParser as configparser
from contextlib import contextmanager
import datetime
import difflib
import functools
import glob
import logging
import os
import pkg_resources
import re
from select import select
import sys
import subprocess

import psycopg2
from psycopg2.extensions import POLL_OK, POLL_READ, POLL_WRITE


logger = logging.getLogger('dbmigrator')
logger.setLevel(logging.INFO)

handler = logging.StreamHandler(sys.stdout)
handler.setFormatter(logging.Formatter(
    '[%(levelname)s] %(name)s (%(filename)s) - %(message)s'))
logger.addHandler(handler)


_settings = {}


def get_settings():
    return _settings


def set_settings(settings):
    global _settings
    _settings = settings


@contextmanager
def db_connect(*args, **kwargs):
    db_conn = psycopg2.connect(*args, **kwargs)
    try:
        with db_conn:
            yield db_conn
    finally:
        db_conn.close()


@contextmanager
def super_user():
    settings = get_settings()
    super_user = settings.get('super_user', 'postgres')
    with db_connect(settings['db_connection_string'],
                    user=super_user) as db_conn:
        with db_conn.cursor() as cursor:
            yield cursor


def deferred(func):
    """A decorator to mark the migration as deferred."""
    func.dbmigrator_deferred = True
    return func


# psycopg2 / libpq doesn't respond to SIGINT (ctrl-c):
# https://github.com/psycopg/psycopg2/issues/333
# To get around this problem, using code from:
# http://initd.org/psycopg/articles/2014/07/20/cancelling-postgresql-statements-python/ # noqa
def wait_select_inter(conn):
    while 1:
        try:
            state = conn.poll()
            if state == POLL_OK:
                break
            elif state == POLL_READ:
                select([conn.fileno()], [], [])
            elif state == POLL_WRITE:
                select([], [conn.fileno()], [])
            else:
                raise conn.OperationalError(
                    'bad state from poll: {}'.format(state))
        except KeyboardInterrupt:
            conn.cancel()
            continue


def get_settings_from_entry_points(settings, contexts):
    context_settings = {}

    for context in contexts:
        entry_points = pkg_resources.get_entry_map(
            context, __package__).values()
        for entry_point in entry_points:
            setting_name = entry_point.name
            if settings.get(setting_name):
                # don't overwrite settings given from the CLI
                continue

            value = entry_point.load()
            if callable(value):
                value = value()

            old_value = context_settings.get(setting_name)
            if (old_value and old_value != value or
                    isinstance(old_value, list) and value not in old_value):
                if not isinstance(old_value, list):
                    context_settings[setting_name] = [old_value]
                context_settings[setting_name].append(value)
            else:
                context_settings[setting_name] = value

    for name, value in context_settings.items():
        settings[name] = value


def get_settings_from_config(filename, config_names, settings):
    config = configparser.ConfigParser()
    config.read(filename)
    for name in config_names:
        setting_name = name.replace('-', '_')
        if settings.get(setting_name):
            # don't overwrite settings given from the CLI
            continue

        for section_name in config.sections():
            try:
                value = config.get(section_name, name)
                settings[setting_name] = value
                break
            except configparser.NoOptionError:
                pass


def with_cursor(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        if not kwargs.get('db_connection_string'):
            raise Exception('db-connection-string missing')
        db_conn_str = kwargs.get('db_connection_string')
        with db_connect(db_conn_str) as db_conn:
            with db_conn.cursor() as cursor:
                return func(cursor, *args, **kwargs)
    return wrapper


def import_migration(path):
    dirname, basename = os.path.split(path)
    if dirname not in sys.path:
        sys.path.append(dirname)
    module_name = basename.rsplit('.', 1)[0]
    return __import__(module_name)


def get_migrations(migration_directories, import_modules=False, reverse=False):
    paths = [os.path.join(md, '*.py') for md in migration_directories]
    python_files = functools.reduce(
        lambda a, b: a + b,
        [glob.glob(path) for path in paths], [])
    for path in sorted(python_files,
                       key=lambda path: os.path.basename(path),
                       reverse=reverse):
        filename = os.path.basename(path)
        m = re.match('([0-9]+)_(.+).py$', filename)
        if m:
            version, migration_name = m.groups()
            if import_modules:
                yield version, migration_name, import_migration(path)
            else:
                yield version, migration_name


def get_schema_versions(cursor, versions_only=True, raise_error=True,
                        include_deferred=True, order_by='version'):
    try:
        cursor.execute('SELECT * FROM schema_migrations ORDER BY {}'
                       .format(order_by))
        for i in cursor.fetchall():
            if not include_deferred and i[1] is None:
                continue
            if versions_only:
                yield i[0]
            else:
                yield i
    except psycopg2.ProgrammingError as e:
        if raise_error:
            raise
        logger.warning(str(e))
        logger.warning('You may need to create the schema_migrations table: '
                       'dbmigrator init --help')


def get_pending_migrations(migration_directories, cursor, import_modules=False,
                           up_to_version=None, include_defers=False):
    migrated_versions = {i[0]: i[1]
                         for i in get_schema_versions(
                             cursor, versions_only=False)}

    migrations = list(get_migrations(migration_directories,
                                     import_modules=True))
    versions = [m[0] for m in migrations]
    if up_to_version:
        try:
            migrations = migrations[:versions.index(up_to_version) + 1]
        except ValueError:
            raise Exception('Version "{}" not found'.format(up_to_version))

    for migration in migrations:
        version, migration_name, mod = migration
        if not import_modules:
            migration = migration[:-1]
        deferred = is_deferred(version, mod, migrated_versions)
        if not deferred or include_defers:
            if hasattr(mod, 'should_run'):
                # repeat migrations are always included
                yield migration
            elif deferred and include_defers or \
                    not deferred and not migrated_versions.get(version):
                yield migration


def compare_schema(db_connection_string, callback, *args, **kwargs):
    old_schema = subprocess.check_output(
        ['pg_dump', '-s', db_connection_string]).decode('utf-8')
    callback(*args, **kwargs)
    new_schema = subprocess.check_output(
        ['pg_dump', '-s', db_connection_string]).decode('utf-8')
    logger.info(''.join(list(
        difflib.unified_diff(old_schema.splitlines(True),
                             new_schema.splitlines(True),
                             n=10))).encode('utf-8'))


def run_migration(cursor, version, migration_name, migration,
                  run_deferred=False):
    if not run_deferred:
        migrated_versions = dict(list(
            get_schema_versions(cursor, versions_only=False,
                                raise_error=False)))
        if is_deferred(version, migration, migrated_versions):
            logger.info('Skipping deferred migration {} {}'
                        .format(version, migration_name))
            return

    try:
        cursor.execute('SAVEPOINT pre_should_run')
        if not migration.should_run(cursor):
            logger.info('Skipping migration {} {}: should_run is false'
                        .format(version, migration_name))
            cursor.execute('ROLLBACK TO pre_should_run')
            cursor.execute('RELEASE SAVEPOINT pre_should_run')
            return
    except AttributeError:
        # not a repeat migration
        pass

    logger.info('Running migration {} {}'.format(version, migration_name))
    migration.up(cursor)
    mark_migration(cursor, version, True)
    cursor.connection.commit()


def rollback_migration(cursor, version, migration_name, migration):
    logger.info('Rolling back migration {} {}'.format(version, migration_name))
    migration.down(cursor)
    mark_migration(cursor, version, False)
    cursor.connection.commit()


def timestamp():
    now = datetime.datetime.utcnow()
    return now.strftime('%Y%m%d%H%M%S')


def mark_migration(cursor, version, completed):
    if completed == 'deferred':
        cursor.execute('INSERT INTO schema_migrations (version, applied) '
                       'VALUES (%s, NULL)', (version,))
    elif completed:
        cursor.execute('INSERT INTO schema_migrations VALUES (%s)', (version,))
    else:
        cursor.execute('DELETE FROM schema_migrations WHERE version = %s',
                       (version,))


def is_deferred(version, migration, migrated_versions):
    # migrated versions: version -> applied timestamp
    timestamp = migrated_versions.get(version, '')
    if timestamp == '':
        return hasattr(migration.up, 'dbmigrator_deferred')
    elif timestamp is None:
        return True
    elif hasattr(migration, 'should_run') and \
            hasattr(migration.up, 'dbmigrator_deferred'):
        return True
    return False
